/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** flashrw.c - IOCTL handlers for reading and writing the Flash memory that
**             is attached to the PCI bridge.
**
** ----------------------------------------------------------------------------
*/

#include "device.h"
#include "drvparam.h"
#include "flash.h"
#include "flashrw.h"
#include <ioctl_code.h>
#include "ioctls.h"
#include <admxrc2/flashinf.h>

static OS_STATUS
mapMethodResultToStatus(
  MethodResult res)
{
  switch (res) {
  case MethodResult_Success:
    return OS_STATUS_SUCCESS;
  case MethodResult_Timeout:
    return OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
  case MethodResult_VerifyFailed:
    return OS_MAKE_STATUS(ADMXRC2_FAILED);
  case MethodResult_InvalidLocation:
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  case MethodResult_NotPresent:
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  case MethodResult_NotSupported:
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  case MethodResult_GeneralFailure:
    return OS_MAKE_STATUS(ADMXRC2_FAILED);
  default:
    return OS_MAKE_STATUS(ADMXRC2_UNKNOWN_ERROR);
  }
}

/*
** ---------------------------------------------------------------------
** Exported routines
** ---------------------------------------------------------------------
*/

OS_STATUS
ADMXRC_EraseFlash(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_ERASEFLASH* pIoctl,
  unsigned long inSize)
{
  MethodResult              res;
  OS_STATUS                 status;
  os_uint32                 start, length, limit;
  os_uint32                 tgtStart, tgtLimit;
  os_boolean                bAllow;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  if (pIoctl->in.signature1 != ADMXRC_IOCTLS_ERASEFLASH_SIGNATURE1 ||
      pIoctl->in.signature2 != ADMXRC_IOCTLS_ERASEFLASH_SIGNATURE2) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  start = pIoctl->in.start;
  length = pIoctl->in.length;
  limit = start + length;
  tgtStart = pDevCtx->info.flash.targetArea.start;
  tgtLimit = tgtStart + pDevCtx->info.flash.targetArea.length;
  bAllow = (start >= tgtStart && limit <= tgtLimit);
  switch (pDevCtx->info.cardInfo.boardType) {
  case ADMXRC2_BOARD_ADMXPL:
    if (admxrcDriverParams.model.xpl.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC2_PRO:
    if (admxrcDriverParams.model.xp.unlockFlash) {
      bAllow = TRUE;
    }
    break;
    
  case ADMXRC2_BOARD_ADPXPI:
    if (admxrcDriverParams.model.xpi.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADPEXRC4FX:
    if (admxrcDriverParams.model.xrce4fx.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC4FX:
    if (admxrcDriverParams.model.xrc4fx.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC5LX:
    if (admxrcDriverParams.model.xrc5lx.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADCPXRC4LX:
    if (admxrcDriverParams.model.xrcp4lx.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC5T1:
    if (admxrcDriverParams.model.xrc5t1.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC5T2:  
    if (admxrcDriverParams.model.xrc5t2.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMAMC5A2:
    if (admxrcDriverParams.model.amc5a2.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC5TZ:
    if (admxrcDriverParams.model.xrc5tz.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  default:
    break;
  }
  if (bAllow) {
    res = eraseFlash(pDevCtx, start, length);
    status = mapMethodResultToStatus(res);
  } else {
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
  }
  
  return status;
}

OS_STATUS
ADMXRC_GetFlashInfo(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_GETFLASHINFO* pIoctl,
  unsigned long outSize)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  
  if (pIoctl == NULL || outSize != sizeof(pIoctl->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  
  if (!pDevCtx->info.flash.present) {
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  }
  
  strncpy(pIoctl->out.deviceName, pDevCtx->info.flash.identity.name, 31);
  pIoctl->out.deviceName[31] = '\0';
  pIoctl->out.deviceSize = pDevCtx->info.flash.size;
  pIoctl->out.deviceBlockSize = pDevCtx->info.flash.layout.eraseBlock;
  pIoctl->out.useableStart = pDevCtx->info.flash.targetArea.start;
  pIoctl->out.useableLength = pDevCtx->info.flash.targetArea.length;
  
  pReq->status.length = sizeof(pIoctl->out);

  return status;
}

OS_STATUS
ADMXRC_GetFlashBlockInfo(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_GETFLASHBLOCKINFO* pIoctl,
  unsigned long inSize,
  unsigned long outSize)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  os_uint32 location, blockStart, blockSize;
  os_boolean bBootBlock;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in) || outSize != sizeof(pIoctl->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  if (!pDevCtx->info.flash.present) {
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  }
  
  location = pIoctl->in.location;
  if (location >= pDevCtx->info.flash.size) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  identifyFlashBlock(pDevCtx, location, &blockStart, &blockSize, &bBootBlock);
  pIoctl->out.start = blockStart;
  pIoctl->out.length = blockSize;
  pIoctl->out.flags = 0;
  if (bBootBlock) {
    pIoctl->out.flags |= ADMXRC2_FLASHBLOCKFLAG_BOOT;
  }

  pReq->status.length = sizeof(pIoctl->out);
  
  return status;
}

OS_STATUS
ADMXRC_ReadFlash(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_READFLASH* pIoctl,
  unsigned long inSize)
{
  MethodResult res;
  OS_STATUS status;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  res = readFlash(pDevCtx, pIoctl->in.start, pReq->param.massSize, pReq->param.massBuf);
  status = mapMethodResultToStatus(res);

  return status;
}

OS_STATUS
ADMXRC_WriteFlash(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_WRITEFLASH* pIoctl,
  unsigned long inSize)
{
  MethodResult              res;
  OS_STATUS                 status;
  unsigned long             length;
  os_uint32                 start, limit;
  os_uint32                 tgtStart, tgtLimit;
  os_boolean                bAllow;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  if (pIoctl->in.signature1 != ADMXRC_IOCTLS_WRITEFLASH_SIGNATURE1 ||
      pIoctl->in.signature2 != ADMXRC_IOCTLS_WRITEFLASH_SIGNATURE2) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  length = pIoctl->in.length;
  if (pReq->param.massSize != pIoctl->in.length) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  
  start = pIoctl->in.start;
  limit = start + length;
  tgtStart = pDevCtx->info.flash.targetArea.start;
  tgtLimit = tgtStart + pDevCtx->info.flash.targetArea.length;
  bAllow = (start >= tgtStart && limit <= tgtLimit);
  switch (pDevCtx->info.cardInfo.boardType) {
  case ADMXRC2_BOARD_ADMXPL:
    if (admxrcDriverParams.model.xpl.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXP:
    if (admxrcDriverParams.model.xp.unlockFlash) {
      bAllow = TRUE;
    }
    break;
    
  case ADMXRC2_BOARD_ADPXPI:
    if (admxrcDriverParams.model.xpi.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADPEXRC4FX:
    if (admxrcDriverParams.model.xrce4fx.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC4FX:
    if (admxrcDriverParams.model.xrc4fx.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC5LX:
    if (admxrcDriverParams.model.xrc5lx.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADCPXRC4LX:
    if (admxrcDriverParams.model.xrcp4lx.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC5T1:
    if (admxrcDriverParams.model.xrc5t1.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC5T2:
    if (admxrcDriverParams.model.xrc5t2.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMAMC5A2:
    if (admxrcDriverParams.model.amc5a2.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC5TZ:
    if (admxrcDriverParams.model.xrc5tz.unlockFlash) {
      bAllow = TRUE;
    }
    break;

  default:
    break;
  }
  if (bAllow) {
    res = writeFlash(pDevCtx, start, length, pReq->param.massBuf);
    status = mapMethodResultToStatus(res);
  } else {
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
  }

  return status;
}
