/*
** ADM-XRC Series Driver
**
** (C) Copyright 2008 Alpha Data
**
** thunk.c - Entry point for the generic section of the driver, compatibility
**           IOCTL handler.
**
** These entry points are called by the platform-specific section of the
** driver.
**
** This file implements the thunk layer, so that IOCTLs originating from a
** 32-bit executable work when the kernel (and driver) have been compiled 
** natively 64-bit.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "device_defs.h"
#include "ioctl_code.h"

/* The following header files declare the IOCTL handlers */
#include "configrw.h"
#include "dma.h"
#include "dmadesc.h"
#include "flashrw.h"
#include "info.h"
#include "intwait.h"
#include "lbrst.h"
#include "loadfpga.h"
#include "owner.h"
#include "rw.h"
#include "setclock.h"
#include "spaceconfig.h"

#if OS_IS_BIARCH
OS_STATUS
onIoctlCompat(
  OS_CLIENT_OBJECT* pClObj,
  OS_IO_REQUEST*    req)
{
  OS_DEVICE_OBJECT* pDevObj;
  DeviceContext* pDevCtx;
  ClientContext* pClCtx;
  OS_STATUS status = OS_STATUS_SUCCESS;
  OS_SPINLOCK_FLAGS f;
  os_boolean bHandled;
  os_boolean bReadOnly, bExclusive;

  os_dbgprint(6, ("onIoctlCompat: entered, client=0x%p code=0x%lx(%ld) inSize=0x%lx outSize=0x%lx\n",
		  pClObj,
		  (unsigned long) req->param.code,
		  (unsigned long) req->param.code,
		  (unsigned long) req->param.inSize,
		  (unsigned long) req->param.outSize));

  pDevObj = os_get_device_object(pClObj);
  pDevCtx = os_get_device_context(pDevObj);
  pClCtx = os_get_client_context(pClObj);

  os_acquire_spinlock(&pDevCtx->owner.lock, &f);
  bReadOnly = (pClCtx->coopLevel == CoopLevel_ReadOnly);
  bExclusive = (pClObj == pDevCtx->owner.exclusive);
  os_release_spinlock(&pDevCtx->owner.lock, f);

  bHandled = TRUE; /* Assume it will be handled in the next switch statement */
  switch (req->param.code) {
  case ADMXRC_IOCTL_OPENCARD2:
    /* Thunk not required */
    status = ADMXRC_OpenCard2(pDevCtx, pClObj, req, req->param.buf, req->param.inSize);
    break;

  case ADMXRC_IOCTL_GETBANKINFO:
    {
      ADMXRC_IOCTLS_GETBANKINFO s;
      ADMXRC_IOCTLS_GETBANKINFO_32* s32;
      
      s32 = (ADMXRC_IOCTLS_GETBANKINFO_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.index = s32->in.index;
	status = ADMXRC_GetBankInfo(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.type = (os_thunk_ulong)s.out.type;
	  s32->out.width = (os_thunk_ulong)s.out.width;
	  s32->out.size = (os_thunk_ulong)s.out.size;
	  s32->out.fitted = (os_thunk_int)s.out.fitted;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_GETCARDINFO2:
    {
      ADMXRC_IOCTLS_GETCARDINFO2 s;
      ADMXRC_IOCTLS_GETCARDINFO2_32* s32;
      
      s32 = (ADMXRC_IOCTLS_GETCARDINFO2_32*)req->param.buf;
      if (s32 == NULL || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	status = ADMXRC_GetCardInfo2(pDevCtx, req, &s, sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.cardId = (os_thunk_ulong)s.out.cardId;
	  s32->out.serialNum = s.out.serialNum;
	  s32->out.boardType = (os_thunk_int)s.out.boardType;
	  s32->out.fpgaType = (os_thunk_int)s.out.fpgaType;
	  s32->out.numClock = (os_thunk_ulong)s.out.numClock;
	  s32->out.numDmaChan = (os_thunk_ulong)s.out.numDmaChan;
	  s32->out.numRamBank = (os_thunk_ulong)s.out.numRamBank;
	  s32->out.numSpace = (os_thunk_ulong)s.out.numSpace;
	  s32->out.ramBanksFitted = s.out.ramBanksFitted;
	  s32->out.boardRevision = s.out.boardRevision;
	  s32->out.logicRevision = s.out.logicRevision;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_GETFLASHINFO:
    /* Thunk not required */
    status = ADMXRC_GetFlashInfo(pDevCtx, req, req->param.buf, req->param.outSize);
    break;

  case ADMXRC_IOCTL_GETFLASHBLOCKINFO:
    {
      ADMXRC_IOCTLS_GETFLASHBLOCKINFO s;
      ADMXRC_IOCTLS_GETFLASHBLOCKINFO_32* s32;
      
      s32 = (ADMXRC_IOCTLS_GETFLASHBLOCKINFO_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.location = s32->in.location;
	status = ADMXRC_GetFlashBlockInfo(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.flags = s.out.flags;
	  s32->out.start = s.out.start;
	  s32->out.length = s.out.length;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_GETGEOGRAPHICALINFO:
    {
      ADMXRC_IOCTLS_GETGEOGRAPHICALINFO s;
      ADMXRC_IOCTLS_GETGEOGRAPHICALINFO_32* s32;
      
      s32 = (ADMXRC_IOCTLS_GETGEOGRAPHICALINFO_32*)req->param.buf;
      if (s32 == NULL || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	status = ADMXRC_GetGeographicalInfo(pDevCtx, req, &s, sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.type = (os_thunk_uint)s.out.type;
	  s32->out.info.pci.domain = s.out.info.pci.domain;
	  s32->out.info.pci.bus = s.out.info.pci.bus;
	  s32->out.info.pci.slot = s.out.info.pci.slot;
	  s32->out.info.pci.function = s.out.info.pci.function;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_GETSPACEINFO:
    {
      ADMXRC_IOCTLS_GETSPACEINFO s;
      ADMXRC_IOCTLS_GETSPACEINFO_32* s32;
      
      s32 = (ADMXRC_IOCTLS_GETSPACEINFO_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.index = s32->in.index;
	status = ADMXRC_GetSpaceInfo(pDevCtx, pClCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.virtualBase = os_dethunk_ptr(s.out.virtualBase);
	  s32->out.virtualSize = (os_thunk_ulong)s.out.virtualSize;
	  s32->out.physicalBase = (os_thunk_ulong)s.out.physicalBase;
	  s32->out.localBase = (os_thunk_ulong)s.out.localBase;
	  s32->out.localSize = (os_thunk_ulong)s.out.localSize;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_GETSPACECONFIG:
    {
      ADMXRC_IOCTLS_GETSPACECONFIG s;
      ADMXRC_IOCTLS_GETSPACECONFIG_32* s32;
      
      s32 = (ADMXRC_IOCTLS_GETSPACECONFIG_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.index = s32->in.index;
	status = ADMXRC_GetSpaceConfig(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.flags = (os_thunk_ulong)s.out.flags;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_GETVERSIONINFO:
    /* Thunk not required */
    status = ADMXRC_GetVersionInfo(pDevCtx, req, req->param.buf, req->param.outSize);
    break;

  default:
    /*
    ** The IOCTL wasnt handled. Try the switch statement below,
    ** if the calling thread owns the card.
    */
    bHandled = FALSE;
    break;
  }
  
  if (bHandled) {
    goto finish;
  }

  if (bReadOnly) {
    req->status.status = OS_STATUS_CANNOTSHARE;
    os_complete_request(req);
    return status;
  }

  switch (req->param.code) {
  /* High priority IOCTLs */
  case ADMXRC_IOCTL_DODMA:
    {
      ADMXRC_IOCTLS_DODMA s;
      ADMXRC_IOCTLS_DODMA_32* s32;

      s32 = (ADMXRC_IOCTLS_DODMA_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.dmadesc = s32->in.dmadesc;
	s.in.offset = s32->in.offset;
	s.in.length = s32->in.length;
	s.in.timeout = s32->in.timeout;
	s.in.direction = s32->in.direction;
	s.in.channel = s32->in.channel;
	s.in.local = s32->in.local;
	s.in.mode = s32->in.mode;
	s.in.flags = s32->in.flags;
	s.in.bUseTimeout = s32->in.bUseTimeout;
	status = ADMXRC_DoDMA(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.remaining = (os_thunk_ulong)s.out.remaining;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_DODMA2D:
    {
      ADMXRC_IOCTLS_DODMA2D s;
      ADMXRC_IOCTLS_DODMA2D_32* s32;

      s32 = (ADMXRC_IOCTLS_DODMA2D_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.dmadesc = s32->in.dmadesc;
	s.in.offset = s32->in.offset;
	s.in.length = s32->in.length;
	s.in.width = s32->in.width;
	s.in.skip = s32->in.skip;
	s.in.local_skip = s32->in.local_skip;
	s.in.timeout = s32->in.timeout;
	s.in.direction = s32->in.direction;
	s.in.channel = s32->in.channel;
	s.in.local = s32->in.local;
	s.in.mode = s32->in.mode;
	s.in.flags = s32->in.flags;
	s.in.bUseTimeout = s32->in.bUseTimeout;
	status = ADMXRC_DoDMA2D(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.remaining = (os_thunk_ulong)s.out.remaining;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_DODMAIMMEDIATE:
    {
      ADMXRC_IOCTLS_DODMAIMMEDIATE s;
      ADMXRC_IOCTLS_DODMAIMMEDIATE_32* s32;

      s32 = (ADMXRC_IOCTLS_DODMAIMMEDIATE_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.buffer = os_thunk_ptr(s32->in.buffer);
	s.in.length = s32->in.length;
	s.in.timeout = s32->in.timeout;
	s.in.direction = s32->in.direction;
	s.in.channel = s32->in.channel;
	s.in.local = s32->in.local;
	s.in.mode = s32->in.mode;
	s.in.flags = s32->in.flags;
	s.in.bUseTimeout = s32->in.bUseTimeout;
	status = ADMXRC_DoDMAImmediate(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.remaining = (os_thunk_ulong)s.out.remaining;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_DODMAIMMEDIATE2D:
     {
      ADMXRC_IOCTLS_DODMAIMMEDIATE2D s;
      ADMXRC_IOCTLS_DODMAIMMEDIATE2D_32* s32;

      s32 = (ADMXRC_IOCTLS_DODMAIMMEDIATE2D_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.buffer = os_thunk_ptr(s32->in.buffer);
	s.in.length = s32->in.length;
	s.in.width = s32->in.width;
	s.in.skip = s32->in.skip;
	s.in.local_skip = s32->in.local_skip;
	s.in.timeout = s32->in.timeout;
	s.in.direction = s32->in.direction;
	s.in.channel = s32->in.channel;
	s.in.local = s32->in.local;
	s.in.mode = s32->in.mode;
	s.in.flags = s32->in.flags;
	s.in.bUseTimeout = s32->in.bUseTimeout;
	status = ADMXRC_DoDMAImmediate2D(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.remaining = (os_thunk_ulong)s.out.remaining;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
     break;

  case ADMXRC_IOCTL_WAITFORINTERRUPT:
    {
      ADMXRC_IOCTLS_WAITFORINTERRUPT s;
      ADMXRC_IOCTLS_WAITFORINTERRUPT_32* s32;

      s32 = (ADMXRC_IOCTLS_WAITFORINTERRUPT_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.timeout = s32->in.timeout;
	s.in.flags = s32->in.flags;
	status = ADMXRC_WaitForInterrupt(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.remaining = (os_thunk_ulong)s.out.remaining;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_READ:
    {
      ADMXRC_IOCTLS_READ s;
      ADMXRC_IOCTLS_READ_32* s32;
      
      s32 = (ADMXRC_IOCTLS_READ_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.width = s32->in.width;
	s.in.local = s32->in.local;
	s.in.flags = s32->in.flags;
	status = ADMXRC_Read(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_WRITE:
    {
      ADMXRC_IOCTLS_WRITE s;
      ADMXRC_IOCTLS_WRITE_32* s32;
      
      s32 = (ADMXRC_IOCTLS_WRITE_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.width = s32->in.width;
	s.in.local = s32->in.local;
	s.in.flags = s32->in.flags;
	status = ADMXRC_Write(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  /* Low priority IOCTLs */
  case ADMXRC_IOCTL_PROGRAMBYTES:
    /* No thunk required */
    status = ADMXRC_ProgramBytes(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_READREG:
    /* No thunk required */
    status = ADMXRC_ReadReg(pDevCtx, req, req->param.buf, req->param.inSize, req->param.outSize);
    break;

  case ADMXRC_IOCTL_WRITEREG:
    /* No thunk required */
    status = ADMXRC_WriteReg(pDevCtx, req, req->param.buf, req->param.inSize);
    break;

  case ADMXRC_IOCTL_ERASEFLASH:
    {
      ADMXRC_IOCTLS_ERASEFLASH s;
      ADMXRC_IOCTLS_ERASEFLASH_32* s32;
      
      s32 = (ADMXRC_IOCTLS_ERASEFLASH_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.signature1 = s32->in.signature1;
	s.in.start = s32->in.start;
	s.in.length = s32->in.length;
	s.in.signature2 = s32->in.signature2;
	status = ADMXRC_EraseFlash(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_CANCELWAITFORINTERRUPT:
    status = ADMXRC_CancelWaitForInterrupt(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_REGINTERRUPTEVENT:
    {
      ADMXRC_IOCTLS_REGINTERRUPTEVENT s;
      ADMXRC_IOCTLS_REGINTERRUPTEVENT_32* s32;
      
      s32 = (ADMXRC_IOCTLS_REGINTERRUPTEVENT_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.event = s32->in.event;
	status = ADMXRC_RegInterruptEvent(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_UNREGINTERRUPTEVENT:
    {
      ADMXRC_IOCTLS_UNREGINTERRUPTEVENT s;
      ADMXRC_IOCTLS_UNREGINTERRUPTEVENT_32* s32;
      
      s32 = (ADMXRC_IOCTLS_UNREGINTERRUPTEVENT_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.event = s32->in.event;
	status = ADMXRC_UnregInterruptEvent(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_LOADFPGA:
    /* Thunk not required */
    status = ADMXRC_LoadFPGA(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_LOADFPGADMA:
    {
      ADMXRC_IOCTLS_LOADFPGADMA s;
      ADMXRC_IOCTLS_LOADFPGADMA_32* s32;
      
      s32 = (ADMXRC_IOCTLS_LOADFPGADMA_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.channel = s32->in.channel;
	s.in.flags = s32->in.flags;
	s.in.buffer = os_thunk_ptr(s32->in.buffer);
	s.in.length = s32->in.length;
	status = ADMXRC_LoadFPGADMA(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_LOCALBUSRESET:
    status = ADMXRC_LocalBusReset(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_MAPDIRECTMASTER:
    {
      ADMXRC_IOCTLS_MAPDIRECTMASTER s;
      ADMXRC_IOCTLS_MAPDIRECTMASTER_32* s32;

      s32 = (ADMXRC_IOCTLS_MAPDIRECTMASTER_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.dmadesc = s32->in.dmadesc;
	s.in.offset = s32->in.offset;
	s.in.length = s32->in.length;
	s.in.maxpage = s32->in.maxpage;
	status = ADMXRC_MapDirectMaster(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.pagelen = (os_thunk_ulong)s.out.pagelen;
	  s32->out.pagebits = (os_thunk_ulong)s.out.pagebits;
	  s32->out.pagespan = (os_thunk_ulong)s.out.pagespan;
	  s32->out.bytespan = (os_thunk_ulong)s.out.bytespan;
	  s32->out.initoffs = (os_thunk_ulong)s.out.initoffs;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_READCONFIG:
    {
      ADMXRC_IOCTLS_READCONFIG s;
      ADMXRC_IOCTLS_READCONFIG_32* s32;

      s32 = (ADMXRC_IOCTLS_READCONFIG_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.index = s32->in.index;
	status = ADMXRC_ReadConfig(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.value = s.out.value;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_READFLASH:
    {
      ADMXRC_IOCTLS_READFLASH s;
      ADMXRC_IOCTLS_READFLASH_32* s32;

      s32 = (ADMXRC_IOCTLS_READFLASH_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.start = s32->in.start;
	s.in.length = s32->in.length;
	status = ADMXRC_ReadFlash(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_SETCLOCK2:
    {
      ADMXRC_IOCTLS_SETCLOCK2 s;
      ADMXRC_IOCTLS_SETCLOCK2_32* s32;

      s32 = (ADMXRC_IOCTLS_SETCLOCK2_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.clock_index = s32->in.clock_index;
	s.in.freq = s32->in.freq;
	status = ADMXRC_SetClock2(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.actual = (os_thunk_ulong)s.out.actual;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_SETSPACECONFIG:
    {
      ADMXRC_IOCTLS_SETSPACECONFIG s;
      ADMXRC_IOCTLS_SETSPACECONFIG_32* s32;
      
      s32 = (ADMXRC_IOCTLS_SETSPACECONFIG_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.index = s32->in.index;
	s.in.flags = s32->in.flags;
	status = ADMXRC_SetSpaceConfig(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_SETUPDMA:
    {
      ADMXRC_IOCTLS_SETUPDMA s;
      ADMXRC_IOCTLS_SETUPDMA_32* s32;

      s32 = (ADMXRC_IOCTLS_SETUPDMA_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.buffer = os_thunk_ptr(s32->in.buffer);
	s.in.size = s32->in.size;
	s.in.flags = s32->in.flags;
	status = ADMXRC_SetupDMA(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.dmadesc = (os_thunk_ulong)s.out.dmadesc;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_SETUPDMAREMOTE:
    {
      ADMXRC_IOCTLS_SETUPDMAREMOTE s;
      ADMXRC_IOCTLS_SETUPDMAREMOTE_32* s32;

      s32 = (ADMXRC_IOCTLS_SETUPDMAREMOTE_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in) || req->param.outSize != sizeof(s32->out)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.flags = s32->in.flags;
	s.in.maxPages = s32->in.maxPages;
	s.in.pageLength = s32->in.pageLength;
	s.in.pageBits = s32->in.pageBits;
	s.in.pagesSpanned = s32->in.pagesSpanned;
	s.in.bytesSpanned = s32->in.bytesSpanned;
	s.in.initOffset = s32->in.initOffset;
	status = ADMXRC_SetupDMARemote(pDevCtx, req, &s, sizeof(s.in), sizeof(s.out));
	if (OS_SUCCESS(status)) {
	  s32->out.dmadesc = (os_thunk_ulong)s.out.dmadesc;
	  req->status.length = sizeof(s32->out);
	}
      }
    }
    break;

  case ADMXRC_IOCTL_SYNCDIRECTMASTER:
    {
      ADMXRC_IOCTLS_SYNCDIRECTMASTER s;
      ADMXRC_IOCTLS_SYNCDIRECTMASTER_32* s32;

      s32 = (ADMXRC_IOCTLS_SYNCDIRECTMASTER_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.dmadesc = s32->in.dmadesc;
	s.in.offset = s32->in.offset;
	s.in.length = s32->in.length;
	s.in.syncmode = s32->in.syncmode;
	status = ADMXRC_SyncDirectMaster(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_UNSETUPDMA:
    {
      ADMXRC_IOCTLS_UNSETUPDMA s;
      ADMXRC_IOCTLS_UNSETUPDMA_32* s32;

      s32 = (ADMXRC_IOCTLS_UNSETUPDMA_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.dmadesc = s32->in.dmadesc;
	status = ADMXRC_UnsetupDMA(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_WRITECONFIG:
    {
      ADMXRC_IOCTLS_WRITECONFIG s;
      ADMXRC_IOCTLS_WRITECONFIG_32* s32;

      s32 = (ADMXRC_IOCTLS_WRITECONFIG_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.signature1 = s32->in.signature1;
	s.in.index = s32->in.index;
	s.in.value = s32->in.value;
	s.in.signature2 = s32->in.signature2;
	status = ADMXRC_WriteConfig(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  case ADMXRC_IOCTL_WRITEFLASH:
    {
      ADMXRC_IOCTLS_WRITEFLASH s;
      ADMXRC_IOCTLS_WRITEFLASH_32* s32;

      s32 = (ADMXRC_IOCTLS_WRITEFLASH_32*)req->param.buf;
      if (s32 == NULL || req->param.inSize != sizeof(s32->in)) {
	status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      } else {
	s.in.signature1 = s32->in.signature1;
	s.in.start = s32->in.start;
	s.in.length = s32->in.length;
	s.in.signature2 = s32->in.signature2;
	status = ADMXRC_WriteFlash(pDevCtx, req, &s, sizeof(s.in));
      }
    }
    break;

  default:
    req->status.status = OS_STATUS_INVALIDIOCTL;
    os_complete_request(req);
    return status;
  }

 finish:
  if (!OS_PENDING(status)) {
    req->status.status = status;
    os_complete_request(req);
  }

  return status;
}
#endif
